/************************************************************
 *
 * Connection Controller
 *
 * Performs everything related to synchronization and
 * up- and download of files and other data.
 *
 ************************************************************/

Ext.define('POSync.controller.Connection', {
    extend: 'Ext.app.Controller',
    id: 'Connection',

    debug: true,                                        // show debug messages?
    ignoreP: 0,                                         // tell server to ignore actions?

    /**
     * Main stores to work with during sync
     */
    projectsStore: null,				// List of projects - necessary before anything else
    remoteDirStore: null,				// List of project files on the server side
    localDirStore: null,				// List of project files on the local (client) side
    oldLocalDirStore: null,				// List of project files from last run of POSync
    operationsStore: null,				// List of sync operations to be executed

    refs: [
        { ref: 'panel', selector: '#panel'},
        { ref: 'tabpanel', selector: '#tabpanel'},
        { ref: 'projects', selector: '#panel #projects'},
        { ref: 'files', selector: '#panel #files'},
        { ref: 'operations', selector: '#panel #operations'},
        { ref: 'localDir', selector: '#panel #localdir'},
    ],

    /* This is called before the viewport is created */
    init: function() {
        // console.log ('Connection.init');
        this.control({
        });
    },

    /* This is called after the viewport is created */
    onLaunch: function() {
        var me = this;
        if (me.debug) if (me.debug) console.log ('Connection.onLaunch: controller onLaunch');

        me.projectsStore = me.getProjects().getStore();
        me.remoteDirStore = me.getFiles().getStore();
        me.localDirStore = me.getLocalDir().getStore();
        me.operationsStore = me.getOperations().getStore();

        var app = me.getApplication();
        me.guiController = app.getController('Gui');

        // Tray Icon: receives events when clicking on "sync" menu on Tray icon
        ipcRenderer.on('sync', (event, arg) => {
            if (me.debug) console.log('Connection.onLaunch: sync: '+arg);
            switch (arg) {
            case 'sync':
                me.sync();
                break;
            case 'login':
                me.login();
                break;
            default:
                console.log('Connection.onLaunch: Unknown IPC message from main.js: '+arg);
                alert('Connection.onLaunch: Unknown IPC message from main.js: '+arg);
            }
        });
    },

    /**
     * Convert a time to a PostgreSQL ISO date string
     */
    timeToIso: function(time) {
        var date = new Date(time);
        return date.toISOString().replace("T", " ").replace("Z", "");
    },

    /**
     * Login to the server and get a token that we will use for the
     * following requests.
     */
    login: function (request_options) {
        var me = this;
        if (me.debug) if (me.debug) console.log('Connection.login');

        /* Send the login request */
        var configData = me.guiController.configData;
        if (me.debug) if (me.debug) console.log ('Connection.login: url('+configData.url+') email('+configData.email+') password('+configData.password+')');
        Ext.Ajax.request({
            url: configData.url + '/intranet/auto-login-token',
            params: {
                email: configData.email,
                password: configData.password
            },
            callback: function (options, success, response) {
                var responseJsonString = response.responseText;
                var responseObject = JSON.parse(responseJsonString);

                if (!success) {
                    if (me.debug) if (me.debug) console.log('Connection.login: not successfull');
                    Ext.Msg.alert('Login failed', 'Message from server: "'+responseObject.message+'"');
                    return;
                }


                if (!responseObject) {
                    if (me.debug) if (me.debug) console.log('Connection.login: error parsing responseText='+responseText);
                    // FIXME: Write some message somewhere
                    return;
                }
                var token = responseObject.token;
                if (!token) {
                    if (me.debug) if (me.debug) console.log('Connection.login: no token in responseText='+responseText);
                    // FIXME: Write some message somewhere
                    return;
                }
                if (me.debug) if (me.debug) console.log('Connection.login: token='+token);
                configData.token = token;

                // load the list of projects available to the user
                me.loadProjects();
            }
        });

    },

    /* Send a request to get the list of projects */
    loadProjects: function () {
        var me = this;
        if (me.debug) if (me.debug) console.log('Connection.loadProjects:');

        // Check if already there
        if (me.projectsStore.isLoaded) return;

        var configData = me.guiController.configData;
        var proxy = me.projectsStore.getProxy();
        var baseUrl = configData.url;

        proxy.url = baseUrl+"/intranet-rest/im_project";
        proxy.extraParams = {
            format:             'json',
            query:              "parent_id is NULL and project_nr in ('2016_0049')"     // Select only top-level projects
        };

        me.projectsStore.load({
            callback: function (records,operation,success) {
                if (!success) {
                    if (me.debug) if (me.debug) console.log('Connection.loadProjects: not successfull');
                    // FIXME: Write some message somewhere
                    return;
                }
                me.projectsStore.isLoaded = success;            // mark the store as loaded

                // Sync each project
                me.projectsStore.each(function(project) {
                    var projectId = parseInt(project.get('project_id'));
                    if (me.debug) console.log('Connection.loadProjects: sync project #'+projectId);
                    me.sync(projectId);
                });

            }
        });
    },


    /************************************************************
     * Load remote tree of files from server
     ************************************************************/


    /**
     * Send request to server in order to get remote project files
     */
    loadRemoteDirectory: function (projectId, successCallback) {
        var me = this;
        if (me.debug) if (me.debug) console.log('Connection.loadRemoteDirectory: project_id='+projectId);

        // Skip if already loaded
        //if (''+me.remoteDirStore.isLoaded == ''+projectId) return;

        var configData = me.guiController.configData;
        var proxy = me.remoteDirStore.getProxy();
        var baseUrl = configData.url;
        proxy.url = baseUrl+"/intranet-rest-fs-openacs/file-tree.json";
        proxy.extraParams = {
            format:             'json',
            object_id:          projectId
        };

        me.remoteDirStore.load({
            callback: function (records,operation,success) {
                    if (!success) {
                    if (me.debug) if (me.debug) console.log('Connection.login: not successfull');
                    // FIXME: Write some message somewhere
                    return;
                }
                me.remoteDirStore.isLoaded = projectId;

                var root = me.remoteDirStore.getRootNode();
                root.cascadeBy(function(file) {
                    file.set('expandable', true);
                    file.set('local_p', false);
                });

                // Execute callback
                successCallback && successCallback.call(me, projectId);
            }
        });
    },

    /*
     * Get the local directory data and load the LocalFiles store.
     * directory-tree provides a tree with the following fields:
     */
    loadLocalDirectory: function (projectId) {
        var me = this;
        if (me.debug) if (me.debug) console.log('Connection.loadLocalDirectory: '+projectId);
        if (!me.projectsStore.isLoaded) { alert('Projects not yet loaded'); return; }

        // Get the project path
        var configData = me.guiController.configData;
        var project = me.projectsStore.getById(''+projectId);
        var path = configData.topdir+'/'+project.get('project_path');

        // make sure the path exists
        var fs = require('fs');
        if (!fs.existsSync(configData.topdir)) { fs.mkdirSync(configData.topdir); }
        if (!fs.existsSync(path)) { fs.mkdirSync(path); }

        // Get the contents
        var dirTree = require ('directory-tree');
        var tree = dirTree(path,{attributes:['mode', 'mtime', 'mtimeMs']});
        me.fixLocalTree(tree, projectId);
        me.localDirStore.loadTree (tree);
        me.localDirStore.isLoaded = projectId;
    },

    /**
     * Fix a tree before loading it
     */
    fixLocalTree: function (tree, projectId) {
        var me = this;

        // Get the project path
        var fs = require('fs');
        var crypto = require('crypto');
        tree.local_p = true;
        tree.live_version_id = 0;

        if (tree) {
            if (tree.type == 'directory') {
                tree.type = 'folder';
                tree.expandable = true;
                tree.expanded = tree.children && tree.children.length > 0;
                tree.leaf = false;
                tree.mime_type = '';
            } else {
                var stats = fs.statSync(tree.path);
                var sum = crypto.createHash('sha1');
                sum.update(fs.readFileSync(tree.path));
                var sha1 = sum.digest('hex');
                
                tree.expandable = false;
                tree.expanded = false;
                tree.leaf = true;
                tree.versions = [{
                    version_id: 0,
                    content_length: stats.size,
                    mime_type: tree.type,
                    sha1: sha1,
                    creation_date: me.timeToIso(stats.birthtimeMs),
                    modification_date: me.timeToIso(stats.mtimeMs),
                    description: "",
                    title: "",
                    upload_comment: "",
                    status_id: 0,
                    type_id: 0
                }]
            }
            tree.iconCls = 'icon-' + tree.type;
            if (tree.children) {
                var me = this;
                tree.children.forEach (function(child) {
                    me.fixLocalTree (child);
                });
            }
        }
    },


    /*
     * Save the local directory tree to the "localStorage" persistent
     * storage of the Electron browser.
     */
    saveLocalDirectoryToLocalStorage: function(projectId) {
        var me = this;
        if (me.debug) if (me.debug) console.log('Connection.saveLocalDirectoryToLocalStorage: '+projectId);

        // localStorage.clear();
        var localStorage = window.localStorage;
        for(var storeKey in localStorage) {
            var storeKeyProjectId = storeKey.split('/')[0];
            if (projectId+'' === storeKeyProjectId) { localStorage.removeItem(storeKey); }
        }

        // Go through the list of local files and store
        var root = me.localDirStore.getRootNode();
        root.cascadeBy(function(localFile) {
            me.saveLocalDirectoryFileToLocalStorage(projectId, localFile);           
        });
    },


    /*
     * Save the local file status to the "localStorage" persistent
     * storage of the Electron browser.
     */
    saveLocalDirectoryFileToLocalStorage: function(projectId, localFile) {
        var me = this;
        if (me.debug) if (me.debug) console.log('Connection.saveLocalDirectoryFileToLocalStorage: '+projectId);

        if ("root" == localFile.get('id')) return;		// Exclude root

        var pathArray = localFile.getPathArray();
        var purePathArray = pathArray.slice(0);
        var fileName = localFile.get('name');
        var versions = localFile.get('versions');
        
        var type = localFile.get('type');
        switch (type) {
        case 'file':
            var spacer = ((purePathArray.length == 0) ? '' : '/');
            var storeKey = projectId+'/'+purePathArray.join('/')+spacer+fileName;
            var storeValue = versions[0].sha1;
            localStorage.setItem(storeKey, storeValue);
            break;
        case 'folder':
            var storeKey = projectId+'/'+purePathArray.join('/')+'/';
            var storeValue = "";
            localStorage.setItem(storeKey, storeValue);
            break;
        default:
            alert('saveLocalDirectoryFileToLocalStorage: invalid type for local file: '+type);
        }
    },


    /*
     * Create a path in the me.oldLocalDirStore for a
     * path, so that a file will can be added correctly.
     * Returns the node representing the path.
     */
    loadOldLocalDirectoryCreatePath: function(projectId, filePath) {
        var me = this;
        // if (me.debug) if (me.debug) console.log('Connection.loadOldLocalDirectoryCreatePath: '+projectId+', '+filePath);

        // Loop through the path items
        var node = me.oldLocalDirStore.getRootNode();
        for (var i = 0; i < filePath.length; i++) {
            var pathItem = filePath[i];

            // Loop through the children and check if pathItem exists
            var childNode = null;
            for (var j = 0; j < node.childNodes.length; j++) {
                var childNodeCandidate = node.childNodes[j];
                var childNodeCandidateName = childNodeCandidate.get('name');
                if (childNodeCandidateName === pathItem) {
                    childNode = childNodeCandidate;
                    break;
                }
            }

            // Create the child node if not yet there
            if (!childNode) {
                childNode = Ext.create('POSync.model.File', {
                    id: pathItem,
                    name: pathItem,
                    type: 'folder',
                    local_p: 't',
                    expandable: true,
                    leaf: false
                });
                node.appendChild(childNode);
            }

            node = childNode;
        }

        // if (me.debug) if (me.debug) console.log('Connection.loadOldLocalDirectoryCreatePath: finished');
        return node;
    },


    /*
     * Load the contents of the "localStorage" persistent storage
     * of the Electron browser into a TreeStore.
     */
    loadOldLocalDirectoryFromLocalStorage: function(projectId) {
        var me = this;
        if (me.debug) if (me.debug) console.log('Connection.loadOldLocalDirectoryFromLocalStorage: '+projectId);

        // Setup an empty tree for the old version of files
        me.oldLocalDirStore = Ext.create('POSync.store.Files', {
            autoLoad: false,
            root:{expanded: true}
        });

        // localStorage.clear();
        var localStorage = window.localStorage;
        for(var storeKey in localStorage) {
            var storeValue = localStorage.getItem(storeKey);
            var storeKeyProjectId = storeKey.split('/')[0];
            if (projectId+'' !== storeKeyProjectId) continue;

            var pathArray = storeKey.split('/').slice(1);		// Remove the first part containing projectId
            var fileBody = pathArray[pathArray.length-1];
            var filePathArray = pathArray.slice(0,-1);
            var filePath = filePathArray.join('/');
            if ("" !== fileBody) {								// empty fileBody indicates a folder
                var pathNode = me.loadOldLocalDirectoryCreatePath(projectId, filePathArray);
                var file = Ext.create('POSync.model.File', {
                    id: storeKey,
                    name: fileBody,
                    type: 'file',
                    local_p: 't',
                    expandable: false,
                    leaf: true,
                    file_path: filePath,
                    file_sha1: storeValue
                });
                var versions = [{
                    version_id: 0,
                    sha1: storeValue
                }]
                file.set('versions', versions);
                pathNode.appendChild(file);
            }
        }
    },

    /************************************************************
     * Sync
     ************************************************************/

    /**
     * Sync - Loop through remote files
     * and check if we need to update local files accordingly.
     */
    syncFilesFromRemoteToLocal: function(projectId) {
        var me = this;
        if (me.debug) if (me.debug) console.log('Connection.syncFilesFromRemoteToLocal');
        var configData = me.guiController.configData;

        // Go through the list of remote files and check that a local file exists.
        var root = me.remoteDirStore.getRootNode();
        root.cascadeBy(function(remoteFile) {
            if ("root" == remoteFile.get('id')) return;                // Exclude folders and the root

            var pathArray = remoteFile.getPathArray();
            var purePathArray = pathArray.slice(0);
            var fileName = remoteFile.get('name');
            if (me.ignoreLocalFile(remoteFile))
                return;

            var liveVersionId = remoteFile.get('live_version_id');
            var type = remoteFile.get('type');
            if ("file" == type && (!liveVersionId || 0 == liveVersionId)) return;    // deleted file
            pathArray.push(fileName);
            if (me.debug) if (me.debug) console.log('Connection.sync remote: path='+pathArray);

            switch (type) {
            case 'file':
                var localFile = me.localDirStore.searchUsingPathArray(pathArray);
                var oldLocalFile = me.oldLocalDirStore.searchUsingPathArray(pathArray);
                me.syncFile(projectId, oldLocalFile, localFile, remoteFile);
            case 'folder':
                // always create the folder at the moment
                var localFile = me.localDirStore.searchUsingPathArray(purePathArray);
                if (!localFile) {
                    // Create a new local folder
                    var fs = require('fs');
                    var project = me.projectsStore.getById(''+projectId);
                    var projectPath = configData.topdir+'/'+project.get('project_path');
                    var path = projectPath+'/'+purePathArray.join('/');
                    if (!fs.existsSync(path)) { fs.mkdir(path); }
                }
                break;
            default: alert('syncFilesFromRemoteToLocal: invalid type for remote file: '+type);
            }
        });
    },

    
    /**
     * Ignore certain files?
     * By default ignore ".*~"
     */
    ignoreLocalFile: function(localFile) {
        var me = this;
        var fileName = localFile.get('name');
        // if (me.debug) if (me.debug) console.log('Connection.ignoreLocalFile: '+fileName);

        var ignoreP = false;
        if (/~$/.test(fileName)) ignoreP = true;
        if (/\.bak$/.test(fileName)) ignoreP = true;

        if (me.debug) if (me.debug) console.log('Connection.ignoreLocalFile: '+fileName+': result='+ignoreP);
        return ignoreP;
    },


    /**
     * Sync - Loop through local files
     * and check if we need to upload a local file to the server.
     */
    syncFilesFromLocalToRemote: function(projectId) {
        var me = this;
        if (me.debug) if (me.debug) console.log('Connection.syncFilesFromLocalToRemote');
        var configData = me.guiController.configData;

        // Go through the list of local files and check that a remote file exists
        var root = me.localDirStore.getRootNode();
        root.cascadeBy(function(localFile) {
            if ("root" == localFile.get('id')) return;		// Exclude root
            if (me.ignoreLocalFile(localFile))
                return;
            
            var pathArray = localFile.getPathArray();
            var purePathArray = pathArray.slice(0);
            var fileName = localFile.get('name');
            pathArray.push(fileName);
            if (me.debug) console.log('Connection.sync local: path='+pathArray);
            
            var type = localFile.get('type');
            switch (type) {
            case 'file':
                var oldLocalFile = me.oldLocalDirStore.searchUsingPathArray(pathArray);
                var remoteFile = me.remoteDirStore.searchUsingPathArray(pathArray);
                if (remoteFile) {                    // The remote file might have been deleted completely...
                    // Check if live_version was deleted.
                    var remoteFileLiveVersionId = remoteFile.get('live_version_id');
                    if (0 == remoteFileLiveVersionId || !remoteFileLiveVersionId)
                        remoteFile = null;                       // Remote file live_version was deleted
                }
                me.syncFile(projectId, oldLocalFile, localFile, remoteFile);
                break;
            case 'folder':
                var remoteFile = me.remoteDirStore.searchUsingPathArray(purePathArray);
                if (!remoteFile) {
                    var operation = Ext.create('POSync.model.Operation', {
                        id: 'mkdir_'+pathArray.join('_'),
                        op: 'mkdir',
                        project_id: projectId,
                        file_id: localFile.get('id'),
                        file_path: purePathArray.slice(0,-1).join('/'),
                        file_name: localFile.get('name'),
                        local_file: localFile,
                        remote_file: remoteFile,
                        status: 'created'
                    });
                    me.operationsStore.add(operation);
                    if (me.debug) console.log('Connection.sync local: mkdir: path='+purePathArray.join('/'));
                }
                break;
            default: alert('syncFilesFromLocalToRemote: invalid type for local file: '+type);
            }
        });
    },

    /**
     * Sync - Loop through previous version of local files
     * and check if files were deleted.
     */
    syncFilesFromOldLocalToLocal: function(projectId) {
        var me = this;
        if (me.debug) if (me.debug) console.log('Connection.syncFilesFromOldLocalToLocal');
        var configData = me.guiController.configData;

        // Go through the list of old local files and check that file still exists locally
        var root = me.oldLocalDirStore.getRootNode();
        root.cascadeBy(function(oldLocalFile) {
            if ("root" == oldLocalFile.get('id')) return;                // Exclude folders and the root
            if (me.ignoreLocalFile(oldLocalFile))
                return;

            var pathArray = oldLocalFile.getPathArray();
            var purePathArray = pathArray.slice(0);
            var fileName = oldLocalFile.get('name');
            pathArray.push(fileName);
            if (me.debug) if (me.debug) console.log('Connection.syncFilesOldLocalToLocal: path='+pathArray);

            var type = oldLocalFile.get('type');
            switch (type) {
            case 'file':
                var localFile = me.localDirStore.searchUsingPathArray(pathArray);
                var remoteFile = me.remoteDirStore.searchUsingPathArray(pathArray);
                me.syncFile(projectId, oldLocalFile, localFile, remoteFile);
                break;
            case 'folder':
                // No action on folders at the moment
                break;
            default: alert('syncFilesFromOldLocalToLocal: invalid type for remote file: '+type);
            }
        });
    },

    /**
     * Determine the status of a (local or oldLical) file:
     * "" - unknown - this should only appear if a file has not been loaded yet.
     * "live" - Found SHA1 to be the latest "live" version of the remote file
     * "old" - Found SHA1 in some older version of the remote file
     * "new" - SHA1 not found in version list, possibly because it's new
     */
    syncFileStatus: function(sha1, remoteFile) {
        if ("" === sha1) return "";						// empty SHA1 - return empty string as "unknown"

        var remoteVersions = [];
        var remoteLiveVersionSHA1 = null;
        if (remoteFile) {
            remoteVersions = remoteFile.get('versions');

            // Get the SHA1 of the remote live version
            var remoteLiveVersionId = remoteFile.get('live_version_id');
            for (var i = 0; i < remoteVersions.length; i++) {
                var v = remoteVersions[i];
                var version_id = v.version_id;
                if (version_id == remoteLiveVersionId) { remoteLiveVersion = v; }
            }
            if (!remoteLiveVersion) {
                alert('Connection.syncFile: Did not find the live_version_id='+remoteLiveVersionId+' in version array.');
                return;
            }
            remoteLiveVersionSHA1 = remoteLiveVersion.sha1;
        }

        if (sha1 === remoteLiveVersionSHA1) return "live";	// "live" latest version of the file?

        for (var i = 0; i < remoteVersions.length; i++) {
            if (sha1 === remoteVersions[i].sha1) return "old";	// some older version of the file
        }

        return "new";										// SHA1 is there but doesn't match any existing version
    },

    /**
     * Sync a file.
     * This is called for every local, oldLocal and remote file
     * from the cascade(...) of each TreeStore.
     *
     * The algorithm checks the SHA1 of the tree files in
     * order to determine the version of the file and then
     * takes sync decisions.
     *
     */
    syncFile: function(projectId, oldLocalFile, localFile, remoteFile) {
        var me = this;
        if (me.debug) console.log('Connection.syncFile');

        var actionTable = [
            //Old	Local	Remote	Action		Tested?	Comment
            ["",	"",	"",	"error",	"todo",	"Local and remote unknown, that doesn't work"],
            ["new",	"",	"",	"none",	"",	"done",	"(Backup) file appeared and disappeared quickly - ignore"],
            ["live",	"",	"",	"error",	"todo",	"Local and remote unknown, that doesn't work"],
            ["old",	"",	"",	"error",	"todo",	"Local and remote unknown, that doesn't work"],
            ["",	"new",	"",	"upload",	"todo",	"Local file exist and isn't found remotely"],
            ["new",	"new",	"",	"upload",	"done",	"Local file exist and isn't found remotely"],
            ["live",	"new",	"",	"upload",	"todo",	"Local file exist and isn't found remotely"],
            ["old",	"new",	"",	"upload",	"todo",	"Local file exist and isn't found remotely"],
            ["",	"live",	"",	"error","",	"todo",	"Impossible, remote is necessary to determine local live status"],
            ["new",	"live",	"",	"error",	"todo",	"Impossible, remote is necessary to determine local live status"],
            ["live",	"live",	"",	"error",	"todo",	"Impossible, remote is necessary to determine local live status"],
            ["old",	"live",	"",	"error",	"todo",	"Impossible, remote is necessary to determine local live status"],
            ["",	"old",	"",	"error",	"todo",	"Impossible, remote is necessary to determine local old status"],
            ["new",	"old",	"",	"error",	"todo",	"Impossible, remote is necessary to determine local old status"],
            ["live",	"old",	"",	"error",	"todo",	"Impossible, remote is necessary to determine local old status"],
            ["old",	"old",	"",	"error",	"todo",	"Impossible, remote is necessary to determine local old status"],
            ["",	"",	"live",	"download",	"todo",	"Remote version available, locally nothing there yet"],
            ["new",	"",	"live",	"download",	"todo",	"A local file appeared and disappeared. Get the latest one"],
            ["live",	"",	"live",	"del_remote",	"todo",	"The local live version was deleted - now also delete remotely"],
            ["old",	"",	"live",	"conflict",	"todo",	"An older version was locally deleted, but a newer version exists online"],
            ["",	"new",	"live",	"conflict",	"todo",	"New-born local file conflicts with existing"],
            ["new",	"new",	"live",	"upload",	"done",	"Upload went wrong the last time?"],
            ["live",	"new",	"live",	"upload",	"done",	"The live version was locally modified"],
            ["old",	"new",	"live",	"conflict",	"todo",	"An older version was locally modified, while the a new version is available remotely"],
            ["",	"live",	"live",	"none", 	"todo",	"Local is live version of remote file"],
            ["new",	"live",	"live",	"none", 	"todo",	"Local is live version of remote file"],
            ["live",	"live",	"live",	"none", 	"todo",	"Local is live version of remote file"],
            ["old",	"live",	"live",	"none", 	"todo",	"Local is live version of remote file"],
            ["",	"old",	"live",	"download",	"todo",	"Local SHA1 identified as an older version"],
            ["new",	"old",	"live",	"download",	"todo",	"Local SHA1 identified as an older version"],
            ["live",	"old",	"live",	"download",	"todo",	"Local SHA1 identified as an older version"],
            ["old",	"old",	"live",	"download",	"todo",	"Local SHA1 identified as an older version"],
        ];

        
        // Determine SHA1 and status from files
        var pathArray = null;
        var fileName = null;
        var oldSHA1 = "";
        if (oldLocalFile) {
            var versions = oldLocalFile.get('versions');
            oldSHA1 = versions[0].sha1;
        }
        var oldStatus = me.syncFileStatus(oldSHA1, remoteFile);

        var localSHA1 = "";
        if (localFile) {
            var versions = localFile.get('versions');
            localSHA1 = versions[0].sha1;
            pathArray = localFile.getPathArray();
            fileName = localFile.get('name');
        }
        var localStatus = me.syncFileStatus(localSHA1, remoteFile);

        var remoteSHA1 = "";
        if (remoteFile) {
            var liveVersion = null;
            var versions = remoteFile.get('versions');
            var liveVersionId = remoteFile.get('live_version_id');

            if (!liveVersionId || 0 == liveVersionId) {
                // File with deleted live version.
                // This happens after "deleting" a file using ]po[-Sync.
                // => Just return. So the file will not be in the list of remote files.
                return;
            }
            
            for (var i = 0; i < versions.length; i++) {
                var v = versions[i];
                var version_id = v.version_id;
                if (version_id == liveVersionId) { liveVersion = v; }
            }
            if (!liveVersion) {
                alert('Connection.syncFile: Did not find the live_version_id='+liveVersionId+' in version array.');
                return;
            }
            remoteSHA1 = liveVersion.sha1;
            pathArray = remoteFile.getPathArray();
            fileName = remoteFile.get('name');
        }
        var remoteStatus = me.syncFileStatus(remoteSHA1, remoteFile);

        if (me.debug) console.log('Connection.syncFile: old='+oldStatus+', local='+localStatus+', remote='+remoteStatus);

        if (!pathArray) return;								// new - none - none: no local or remote file...
        var purePathArray = pathArray.slice(0);
        pathArray.push(fileName);
        var fileNameWithPath = '/'+projectId+'/'+pathArray.join('/');
        if (me.debug) console.log('Connection.syncFile: pathArray='+pathArray);


        var action = null;
        var testingStatus = null;
        var comment = null;
        for (var i = 0; i < actionTable.length; i++) {
            var actionLine = actionTable[i];
            if (oldStatus === actionLine[0] && localStatus === actionLine[1] && remoteStatus === actionLine[2]) {
                action = actionLine[3];
                testingStatus = actionLine[4];
                comment = actionLine[5];
            }
        }

        if (!action) {
            action = "none";
            comment = "Didn't find action for status combination: old="+oldStatus+', local='+localStatus+', remote='+remoteStatus;
        }


        if (false) {
            var a = null;								// action - not set yet
            if (!a && "live" === localStatus)					// up to date, independent of oldStatus
                a = "none";
            if (!a && "old" === localStatus)					// older file locally => download
                a = "download";
            
            if (!a && "new" === localStatus && "old" === oldStatus)			// older file modified locally
                a = "conflict";
            if (!a && "new" === localStatus && "" === oldStatus)			// new-born local file conflicts new remote
                a = "conflict";
            if (!a && "new" === localStatus)					// locally modified, conflict already handled
                a = "upload";
            
            if (!a && "" === localStatus && "live" === oldStatus)			// live version locally deleted
                a = "delete_remote";
            if (!a && "" === localStatus && "old" === oldStatus)			// old version deleted locally, newer version is on-line
                a = "conflict";
            if (!a && "" === localStatus)						// try to download
                a = "download";
            
            if (!a)                                                                 // somehow not cought...
                a = "error"
        }
            
        // Check if there was a previous operation in the operationsHash
        var lastOp = me.operationsHash[fileNameWithPath];
        if (lastOp) {
            // We already checked this file from another "angle"
            if (lastOp !== action) {
                alert('Connection.syncFile: found inconsistent result from previous operation: cur='+action+', prev='+lastOp);
            }
            return;									// already successfully processed
        }

        if ("error" === action) {
            alert("Connection: found invalide operation: "+a);
            return;
        }
        
        // Write operation to hash to avoid duplicates
        me.operationsHash[fileNameWithPath] = action;

        // !!! if ("none" === action) return;							// No operation...

        var operation = Ext.create('POSync.model.Operation', {
            id: action+'_'+pathArray.join('_'),
            op: action,
            project_id: projectId,
            file_id: ((remoteFile) ? remoteFile.get('id') : ""),
            file_path: purePathArray.join('/'),
            file_name: fileName,
            file_sha1: localSHA1,
            local_file: localFile,
            remote_file: remoteFile,
            comment: comment,
            err_msg: 'Status: old='+oldStatus+', local='+localStatus+', remote='+remoteStatus,
            status: 'created'
        });
        me.operationsStore.add(operation);
    },

    /**
     * Start the synchronization
     * by loading the files store from the server.
     *
     * Calls syncFilesLoadedCallback to continue
     * after the remote files have arrived
     */
    sync: function (projectId) {
        var me = this;
        if (me.debug) if (me.debug) console.log('Connection.sync: projectId='+projectId);

        // Load the list of local files needed for sync
        me.loadLocalDirectory(projectId);

        // Get the last saved
        me.loadOldLocalDirectoryFromLocalStorage(projectId);

        // Load the list of files and continue with syncCallback
        me.loadRemoteDirectory(projectId, me.syncFilesLoadedCallback);

        // Show "Login" and "Sync" tabs, but only "Sync" button. "Sync" tab is active
        me.guiController.tabsAndButtons(['login', 'sync', 'projects'], ['sync'], 'sync');
        
    },


    /**
     * Sync all loaded projects
     */
    syncAll: function () {
        var me = this;
        if (me.debug) if (me.debug) console.log('Connection.syncAll: starting');

        me.projectsStore.each(function(proj) {
            var id = proj.get('id');
            me.sync(id);
        });
        if (me.debug) if (me.debug) console.log('Connection.syncAll: finished');
    },



    
    /**
     * Sync - continuation after the remote files have arrived.
     * Now we can start with the actual sync
     */
    syncFilesLoadedCallback: function(projectId) {
        var me = this;
        if (me.debug) if (me.debug) console.log('Connection.syncFilesLoadedCallback: projectId='+projectId);
        var configData = me.guiController.configData;

        // Generate sync "operations" (=actions)
        me.operationsHash = {};						// list of operations for eliminating duplicates
        me.operationsStore.removeAll();
        me.syncFilesFromRemoteToLocal(projectId);
        me.syncFilesFromLocalToRemote(projectId);
        me.syncFilesFromOldLocalToLocal(projectId);

        // Execute operations
        me.syncExecOperations(projectId);

        // Don't save the status just like this anymore.
        // Instead, for idempotence, we save only if we executed the operation
        // me.saveLocalDirectoryToLocalStorage(projectId);
    },


    /************************************************************
     * Sync - Exec Operations
     ************************************************************/

    /**
     * Takes a store with sync "Operations" and executes them.
     * This performs the actual sync.
     */
    syncExecOperations: function() {
        var me = this;
        if (me.debug) console.log('Connection.syncExecOperations');

        // Execute all "easy" actions
        // !!! me.syncExecOperationsEasy();

        // Check if there are operations left
        var storeCount = me.operationsStore.count();
        if (storeCount == 0) {
            if (me.debug) console.log('Connection.syncExecOperations: Successfully processed sync, only easy ops');
            return;
        }

        // There are non-easy operations left: Ask for Commit meesage and approval
        me.guiController.tabsAndButtons('sync', 'commit', 'sync');
        var commitText = Ext.getCmp('textfield_commit');
        commitText.setValue("");
        commitText.show();
        // Now only the "Commit" button is active.
        // After pressing it, execution flow continues with syncExecOperationsNonEasy
    },


    /**
     * Executes those actions that don't need user approval
     */
    syncExecOperationsEasy: function() {
        var me = this;
        if (me.debug) console.log('Connection.syncExecOperationsEasy');

        me.operationsStore.each(function(op) {
            if (!op) { alert('syncExecOperationsEasy: found invalid op'); }
            var operation = op.get('op');
            var localFile = op.get('local_file');
            var projectId = parseInt(op.get('project_id'));
            switch (operation) {
            case 'none':
                if (me.debug) console.log('Connection.syncExecOperationsEasy: op='+operation+', id='+op.get('id'));
                break
            case 'download':
                if (me.debug) console.log('Connection.syncExecOperationsEasy: op='+operation+', id='+op.get('id'));
                me.syncDownload(projectId, op);

                // save to local dir only if processed
                if (localFile)
                    me.saveLocalDirectoryFileToLocalStorage(projectId, localFile);
                break
            }
        });

    },


    /**
     * Executes those actions needed user approval
     */
    syncExecOperationsNonEasy: function() {
        var me = this;
        if (me.debug) console.log('Connection.syncExecOperationsNonEasy');


        var commitText = Ext.getCmp('textfield_commit');
        if (1 || commitText.getValue().length > 5) {
            var loginButton = Ext.getCmp('button_login');
            var syncButton = Ext.getCmp('button_sync');
            var commitText = Ext.getCmp('textfield_commit');
            var commitButton = Ext.getCmp('button_commit');
            loginButton.hide();
            syncButton.show();
            commitText.hide();
            commitButton.hide();
        } else {
            alert('Commit message needs to have at least 5 characters');
            return;
        }
        
        // Execute all other actions
        var skip = false;
        me.operationsStore.each(function(op) {
            if (!op) {
                return;
                alert('syncExecOperationsEasy: found invalid op');
            }
            if (skip) return;

            var operation = op.get('op');
            var localFile = op.get('local_file');
            var projectId = parseInt(op.get('project_id'));
            if (me.debug) console.log('Connection.syncExecOperationsNonEasy: op='+operation+', id='+op.get('id'));
            switch (operation) {
            case 'upload':
                me.syncUpload(projectId, op);
                break
            case 'mkdir':
                me.syncMkdirRemote(projectId, op);
                break
            case 'delete_remote':
                me.syncDeleteRemote(projectId, op);
                break
            default:
                alert('syncExecOperationsNonEasy: Internal error: unknown operation='+operation);
            }

            if (localFile)
                me.saveLocalDirectoryFileToLocalStorage(projectId, localFile);

        });

        // Show tabs "Sync" and "Remote", plus "Sync" button. "Sync" tab is active
        me.guiController.tabsAndButtons(['sync', 'remote_files'], 'sync', 'remote_files');
    },

    
    /**
     * Download a remote file to the local file-system.
     *
     * Called from syncExecOperations, which in turn
     * loops through the list of operations created during
     * sync.
     */
    syncDownload: function(projectId, op) {
        var me = this;

        var configData = me.guiController.configData;
        var fileId = op.get('file_id');
        var filePath = op.get('file_path');
        var fileName = op.get('file_name');
        var mimeType = op.get('mime_type');
        if (me.debug) console.log('Connection.syncDownload: path='+filePath+', name='+fileName);

        // Get the project path
        var project = me.projectsStore.getById(''+projectId);
        var projectPath = configData.topdir+'/'+project.get('project_path');
        var path = projectPath+'/'+filePath;

        // make sure the path exists
        var fs = require('fs');
        if (!fs.existsSync(path)) { fs.mkdir(path); }

        var requestUrl = configData.url + '/intranet-rest-fs-openacs/download?file_id='+fileId;
        var xhr = new XMLHttpRequest();
        xhr.open("GET", requestUrl);
        xhr.responseType = "arraybuffer";
        // xhr.responseType = "blob";
        xhr.onload = function () {
            // Asynchronously saving files after they have been downloaded
            if (this.status === 200) {
                if (me.debug) console.log('Connection.syncDownload: successful download - saving');

                // Decode the response encoded by server in base64
                var responseBuffer = xhr.response;
                var buf64 = new Buffer(responseBuffer, 'base64'); // decode

                // make sure the path exists and write the file
                var fs = require('fs');
                if (!fs.existsSync(path)) { fs.mkdir(path); }
                fs.writeFile(path+"/"+fileName, buf64, function() {});

                op.set('status', 'done');

                if (me.debug) console.log('Connection.syncDownload: written file='+path+'/'+fileName);
            } else {
                var responseJsonString = this.responseText;
                var responseObject = JSON.parse(responseJsonString);
                op.set('err_msg', responseObject.message);
                if (me.debug) console.log('Connection.syncDownload: download status='+this.status);
            }
            me.operationsStore.filterBy(me.operationsStore.myFilter);           
        };
        xhr.send();
    },

    /**
     * Upload a local file to the remote file-storage
     */
    syncUpload: function(projectId, op) {
        var me = this;

        var configData = me.guiController.configData;
        var filePath = op.get('file_path');
        var fileName = op.get('file_name');
        var mimeType = op.get('mime_type');
        var localFile = op.get('local_file');
        var localVersions = localFile.get('versions');
        var localVersion = localVersions[0];
        if (me.debug) console.log('Connection.syncUpload: path='+filePath+', name='+fileName);

        // Get the project path
        var project = me.projectsStore.getById(''+projectId);
        var projectPath = configData.topdir+'/'+project.get('project_path');
        var path = projectPath+'/'+filePath;
        var mimeType = 'text/xml';

        // Commit comment
        var commitTextField = Ext.getCmp('textfield_commit');
        var commitComment = commitTextField.getValue();
        
        // make sure the path exists
        var fs = require('fs');
        var fileContent = fs.readFileSync(path+"/"+fileName);

        // POST the file to the upload page
        var xhr = new XMLHttpRequest();
        xhr.operationId = op.get('id');
        var url = configData.url + '/intranet-rest-fs-openacs/upload';
        var formData = new FormData();
        formData.append('project_id', projectId);
        formData.append('path', filePath);
        formData.append('creation_date', localVersion.creation_date);
        formData.append('modification_date', localVersion.modification_date);
        formData.append('sha1', localVersion.sha1);
        formData.append('upload_comment', commitComment);
        formData.append('ignore_p', me.ignoreP);        // Debugging: Don't really execute on server-side...
        var blob = new Blob([fileContent], { type: mimeType});
        formData.append('upload_file', blob, fileName);
        op.set('status', 'running');
        xhr.open("POST", url, true);
        xhr.onload = function() {
            if (this.status === 200) {
                if (me.debug) console.log('Connection.syncUpload: successful upload');
                op.set('status', 'done');
            } else {
                if (me.debug) console.log('Connection.syncUpload: upload status='+this.status);
                op.set('status', 'error');
                var responseJsonString = this.responseText;
                var responseObject = JSON.parse(responseJsonString);
                op.set('err_msg', responseObject.message);
            }
            me.operationsStore.filterBy(me.operationsStore.myFilter);           
        };
        xhr.send(formData);
    },

    /**
     * Create a new folder in the remote file-storage
     */
    syncMkdirRemote: function(projectId, op) {
        var me = this;
        var configData = me.guiController.configData;

        // POST the file to the mkdir page
        var xhr = new XMLHttpRequest();
        var url = configData.url + '/intranet-rest-fs-openacs/mkdir';
        var formData = new FormData();
        formData.append('project_id', projectId);
        formData.append('path', op.get('file_path'));
        formData.append('name', op.get('file_name'));
        formData.append('ignore_p', me.ignoreP);        // Debugging: Don't really execute on server-side...
        xhr.open("POST", url, true);
        xhr.onload = function() {
            if (this.status === 200) {
                if (me.debug) console.log('Connection.syncUpload: successful upload');
                op.set('status', 'done');
            } else {
                if (me.debug) console.log('Connection.syncUpload: upload status='+this.status+': '+this.responseText);
                op.set('status', 'error');
                var responseJsonString = this.responseText;
                var responseObject = JSON.parse(responseJsonString);
                op.set('err_msg', responseObject.message);
            }
            me.operationsStore.filterBy(me.operationsStore.myFilter);           
        };
        xhr.send(formData);
    },

    /**
     * DeleteRemote a local file to the remote file-storage
     */
    syncDeleteRemote: function(projectId, op) {
        var me = this;

        var configData = me.guiController.configData;
        var fileId = op.get('file_id');
        var commitTextField = Ext.getCmp('textfield_commit');
        var commitComment = commitTextField.getValue();
        
        // POST the file to the delete page
        var xhr = new XMLHttpRequest();
        xhr.operationId = op.get('id');
        var url = configData.url + '/intranet-rest-fs-openacs/delete';
        var formData = new FormData();
        formData.append('file_id', fileId);
        formData.append('project_id', projectId);
        formData.append('delete_comment', commitComment);
        formData.append('ignore_p', me.ignoreP);        // Debugging: Don't really execute on server-side...

        op.set('status', 'running');
        xhr.open("POST", url, true);
        xhr.onload = function() {
            if (this.status === 200) {
                if (me.debug) console.log('Connection.syncDeleteRemote: successful delete');
                op.set('status', 'done');
            } else {
                if (me.debug) console.log('Connection.syncDeleteRemote: delete status='+this.status);
                op.set('status', 'error');
                var responseJsonString = this.responseText;
                var responseObject = JSON.parse(responseJsonString);
                op.set('err_msg', responseObject.message);
            }
            me.operationsStore.filterBy(me.operationsStore.myFilter);           
        };
        xhr.send(formData);
    },


    /* End of POSync.controller.Connection */
});
